﻿-----------------------------------------------------------------------------
--
-- data
--   \gain
--     \<playerName>
--       - count, amount
--       \gainDetails
--         \gainName 
--			- count, amount, type
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentGain")
L:RegisterTranslations("enUS", function() return {
	["Gain Recipient List"] = true,
	["Gains"] = true,
	["Gain Details"] = true,
	["Gain Type"] = true,
} end)


L:RegisterTranslations("koKR", function() return {
	["Gain Recipient List"] = "수신자 가능 리스트",
	["Gains"] = "획득",
	["Gain Details"] = "상세 획득",
	["Gain Type"] = "획득 유형",
} end)
-----------------------------------------------------------------------------

local AssessmentGain = Assessment:NewModule("Gain", "AceEvent-2.0", "Parser-3.0")
AssessmentGain.db = Assessment:AcquireDBNamespace("Gain")
Assessment:RegisterDefaults("Gain", "profile", {
	enable	= true,
	
	manaGainColor		= { 0.28, 0.52, 0.84, 1},
	energyGainColor		= {    1, 0.86, 0.10, 1},
	healthGainColor		= { 0.10, 0.89, 0.10, 1},
	rageGainColor		= { 0.89, 0.18, 0.29, 1},
	focusGainColor		= {    1, 0.82,    0, 1},
	happinessGainColor	= {    0,    1,    1, 1},
})

-----------------------------------------------------------------------------

local profile

local emptyMethod = Assessment.EmptyMethod
local new = Assessment.NewTable
local pairs = pairs

-----------------------------------------------------------------------------

Assessment.TagData.GainAmount =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local gain = fightData.gain
				if not gain then return 0 end
				
				local playerGain = gain[data.key]
				if not playerGain then return 0 end

				return playerGain.amount
			 end,
	data = "gain",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.GainCount =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local gain = fightData.gain
				if not gain then return 0 end
				
				local playerGain = gain[data.key]
				if not playerGain then return 0 end

				return playerGain.count
			 end,
	data = "gain",
	guaranteeResult = true,
	numeric = true,
}

-----------------------------------------------------------------------------

Assessment.ListType.GainRecipients =
{
	method = function(fightData)
				if not fightData or not fightData.gain then return emptyMethod end
				return pairs(fightData.gain)
			 end,
	data = "gain",
	description = L["Gain Recipient List"],
}

-----------------------------------------------------------------------------

local gainTypeColorParameterName =
{
	Mana		= "manaGainColor",
	Energy		= "energyGainColor",
	Health		= "healthGainColor",
	Focus		= "focusGainColor",
	Rage		= "rageGainColor",
	Happiness	= "happinessGainColor",
}

Assessment.ColorType.GainType = 
{
	method = function(fightData, data)
				return profile[gainTypeColorParameterName[data.value.type]]
			 end,
	description = L["Gain Type"]
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.GainDetails =
{
	titleTag			= ("%s: [WindowParameter(1)]"):format(L["Gain Details"]),
	listType			= "EnumerateTable",
	listDataTag			= "gain,[WindowParameter(1)],gainDetails",
	barValueTag			= "Amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { count = 2, { tag = "[Rank]. [Key] ([Count])" }, { tag = "[BarValue]" } },
	description			= L["Gain Details"],
	selectable			= false,
	colorTypeList		= { "GainType", count = 1 },
	defaultColor		= { 0.8, 0.8, 0.5 },
	favorite			= false,
}

Assessment.DefaultDisplaySets.Gain =
{
	titleTag			= L["Gains"],
	listType			= "GroupMembers", 
	barValueTag			= "GainAmount", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 1,
	text				= { count = 2, { tag = "[Rank]. [Key] ([GainCount])" }, { tag = "[BarValue]" } },
	description			= L["Gains"],
	defaultColor		= { 0.8, 0.8, 0.5 },
	colorTypeList		= { "Class", count = 1 },
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Gain Details"],
	onClickData			= "[Key]",
}

-----------------------------------------------------------------------------

Assessment.ListType.EnumerateTable.updateData.gain = "gain"

-----------------------------------------------------------------------------

function AssessmentGain:OnInitialize()
	profile = self.db.profile
end

function AssessmentGain:OnEnable()
	self:RegisterParserEvent({ eventType = "Gain" }, "OnParserGainEvent")
	self:RegisterParserEvent({ eventType = "Drain", sourceName_not = false }, "OnParserDrainEvent")
end

function AssessmentGain:OnDisable()
end

-----------------------------------------------------------------------------

function AssessmentGain:OnParserGainEvent(eventData)
	local time = GetTime()

	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.gain = time
		local gain = data.gain
		if not gain then
			gain = new()
			data.gain = gain
		end

		local recipientName = eventData.recipientName
		local playerGain = gain[recipientName]
		local gainDetails
		if playerGain then
			playerGain.amount = playerGain.amount + eventData.amount
			playerGain.count = playerGain.count + 1
			gainDetails = playerGain.gainDetails
		else
			playerGain = new()
			playerGain.count = 1
			playerGain.amount = eventData.amount

			gain[recipientName] = playerGain
			
			gainDetails = new()
			playerGain.gainDetails = gainDetails
		end

		local abilityName = eventData.abilityName
		local gainData = gainDetails[abilityName]
		if gainData then
			gainData.count = gainData.count + 1
			gainData.amount = gainData.amount + eventData.amount
		else
			gainData = new()
			gainData.count = 1
			gainData.amount = eventData.amount
			gainData.type = eventData.attribute
			gainDetails[abilityName] = gainData
		end
	end
end

function AssessmentGain:OnParserDrainEvent(eventData)
	local sourceName = eventData.sourceName
	local time = GetTime()

	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.gain = time
		local gain = data.gain
		if not gain then
			gain = new()
			data.gain = gain
		end

		local playerGain = gain[sourceName]
		local gainDetails
		if playerGain then
			playerGain.amount = playerGain.amount + eventData.amount
			playerGain.count = playerGain.count + 1
			gainDetails = playerGain.gainDetails
		else
			playerGain = new()
			playerGain.count = 1
			playerGain.amount = eventData.amount

			gain[sourceName] = playerGain
			
			gainDetails = new()
			playerGain.gainDetails = gainDetails
		end

		local abilityName = eventData.abilityName
		local gainData = gainDetails[abilityName]
		if gainData then
			gainData.count = gainData.count + 1
			gainData.amount = gainData.amount + eventData.amount
		else
			gainData = new()
			gainData.count = 1
			gainData.amount = eventData.amount
			gainData.type = eventData.attribute
			gainDetails[abilityName] = gainData
		end
	end
end

-----------------------------------------------------------------------------
